import json
from dataclasses import replace
from email.message import Message
from functools import cached_property
from http.client import HTTPResponse
from pathlib import Path
from ssl import create_default_context
from typing import (
    IO,
    Callable,
    List,
    Literal,
    Mapping,
    NoReturn,
    Optional,
    Sequence,
    TypedDict,
    Union,
    cast,
    overload,
)
from urllib.error import HTTPError
from urllib.parse import urljoin
from urllib.request import (
    AbstractHTTPHandler,
    BaseHandler,
    HTTPDefaultErrorHandler,
    HTTPSHandler,
    Request,
    build_opener,
)

from typing_extensions import TypeGuard

from ._get_endpoint_path import get_endpoint_path
from ._server_versions import ServerVersions
from .fetch_json import JSON_CONTENT_TYPE, JsonResponse, fetch_json
from .http_method import HttpMethod
from .path import local_to_absolute_path


class _ErrorChainItem(TypedDict):
    message: str
    type: str


class _CurrentActiveViamJsonHttpErrorBody(TypedDict):
    errorChain: Sequence[_ErrorChainItem]
    stackTrace: str


class _LegacyActiveViamJsonHttpErrorBody(TypedDict):  # ActivePivot < 6.0.0-M1.
    error: _CurrentActiveViamJsonHttpErrorBody
    status: Literal["error"]


_ActiveViamJsonHttpErrorBody = Union[
    _CurrentActiveViamJsonHttpErrorBody, _LegacyActiveViamJsonHttpErrorBody
]


def _is_legacy_activeviam_json_http_error_body(
    body: _ActiveViamJsonHttpErrorBody, /
) -> TypeGuard[_LegacyActiveViamJsonHttpErrorBody]:
    return "status" in body and body.get("status") == "error"


class ActiveViamHttpError(HTTPError):
    def __init__(
        self,
        *,
        body: _CurrentActiveViamJsonHttpErrorBody,
        code: int,
        hdrs: Message,  # spell-checker: disable-line
        url: str,
    ) -> None:
        super().__init__(
            url,
            code,
            body["stackTrace"],
            hdrs,  # spell-checker: disable-line
            None,
        )

        self.error_chain: Sequence[_ErrorChainItem] = body["errorChain"]


class _ActiveViamJsonHttpErrorHandler(HTTPDefaultErrorHandler):
    # Cannot change signature since this is an override.
    def http_error_default(  # pylint: disable=too-many-positional-parameters
        self,
        req: Request,
        fp: Optional[IO[bytes]],
        code: int,
        msg: str,
        hdrs: Message,  # spell-checker: disable-line
    ) -> NoReturn:
        error = HTTPError(
            req.full_url,
            code,
            msg,
            hdrs,  # spell-checker: disable-line
            fp,
        )

        if fp is None or (
            hdrs.get_content_type() != JSON_CONTENT_TYPE  # spell-checker: disable-line
        ):
            raise error

        body: _ActiveViamJsonHttpErrorBody = json.load(fp)
        normalized_body: _CurrentActiveViamJsonHttpErrorBody = (
            body["error"]
            if _is_legacy_activeviam_json_http_error_body(body)
            else cast(_CurrentActiveViamJsonHttpErrorBody, body)
        )
        raise ActiveViamHttpError(
            body=normalized_body,
            code=code,
            hdrs=hdrs,  # spell-checker: disable-line
            url=req.full_url,
        ) from error


_Auth = Callable[[str], Mapping[str, str]]


class _AuthHandler(AbstractHTTPHandler):
    def __init__(self, auth: _Auth, /) -> None:
        super().__init__()

        self._auth = auth

    def _handle_request(self, request: Request) -> Request:
        headers = self._auth(request.full_url)
        request.headers.update(headers)
        return request

    http_request = _handle_request
    https_request = _handle_request


class ActiveViamClient:
    """Used to communicate with ActiveViam servers such as ActivePivot or the Content Server.

    This class uses a custom HTTP error handler to enrich the raised errors with the server stack trace.
    """

    def __init__(
        self,
        url: str,
        /,
        *,
        auth: Optional[_Auth] = None,
        certificate_authority: Optional[Path] = None,
        client_certificate: Optional[Path] = None,
        client_certificate_keyfile: Optional[Path] = None,
        client_certificate_password: Optional[str] = None,
    ) -> None:
        self._url = url.strip("/")

        handlers: List[BaseHandler] = [_ActiveViamJsonHttpErrorHandler()]

        if auth:
            handlers.append(_AuthHandler(auth))

        if certificate_authority or client_certificate:
            context = create_default_context()
            if certificate_authority:
                context.load_verify_locations(
                    cafile=local_to_absolute_path(certificate_authority)
                )
            if client_certificate:
                context.load_cert_chain(
                    certfile=local_to_absolute_path(client_certificate),
                    keyfile=local_to_absolute_path(client_certificate_keyfile)
                    if client_certificate_keyfile
                    else None,
                    password=client_certificate_password,
                )
            handlers.append(HTTPSHandler(context=context))

        self._opener_director = build_opener(*handlers)

    @property
    def url(self) -> str:
        return self._url

    @overload
    def fetch_json(
        self,
        *,
        body: object = ...,
        method: Optional[HttpMethod] = ...,
        namespace: str,
        query: str = ...,
        raw: Literal[True],
        route: str,
    ) -> HTTPResponse:
        ...

    @overload
    def fetch_json(
        self,
        *,
        body: object = ...,
        method: Optional[HttpMethod] = ...,
        namespace: str,
        query: str = ...,
        raw: Literal[False] = ...,
        route: str,
    ) -> JsonResponse:
        ...

    def fetch_json(
        self,
        *,
        body: object = None,
        method: Optional[HttpMethod] = None,
        namespace: str,
        query: str = "",
        raw: bool = False,
        route: str,
    ) -> Union[JsonResponse, HTTPResponse]:
        url = self._get_endpoint_url(namespace=namespace, route=route)

        if query:
            url = f"{url}?{query}"

        response = fetch_json(
            url,
            body=body,
            method=method,
            opener_director=self._opener_director,
            raw=raw,
        )

        if raw:
            return response

        assert isinstance(response, JsonResponse)

        if (
            isinstance(response.body, dict)
            and "data" in response.body
            and response.body.get("status") == "success"
        ):
            # ActivePivot < 6.0.0-M1.
            return replace(response, body=response.body["data"])

        return response

    def ping(self) -> str:
        url = self._get_endpoint_url(namespace="activeviam/pivot", route="ping")
        response = self._opener_director.open(url)
        body = cast(str, response.read().decode("utf8"))
        expected_body = "pong"
        if body != expected_body:
            raise RuntimeError(
                f"Expected `ping()`'s response body to be `{expected_body}` but got `{body}`."
            )
        return body

    def _get_endpoint_path(
        self,
        *,
        attribute_name: Literal["restPath", "wsPath"] = "restPath",
        namespace: str,
        route: str,
    ) -> str:
        return get_endpoint_path(
            attribute_name=attribute_name,
            namespace=namespace,
            route=route,
            server_versions=self.server_versions,
        )

    def _get_url(self, path: str, /) -> str:
        return urljoin(f"{self.url}/", path.lstrip("/"))

    def _get_endpoint_url(
        self,
        *,
        attribute_name: Literal["restPath", "wsPath"] = "restPath",
        namespace: str,
        route: str,
    ) -> str:
        path = self._get_endpoint_path(
            attribute_name=attribute_name, namespace=namespace, route=route
        )
        return self._get_url(path)

    @cached_property
    def server_versions(self) -> ServerVersions:
        url = self._get_url("versions/rest")
        body = fetch_json(url, opener_director=self._opener_director).body
        return cast(ServerVersions, body)
